/*
 *  SnowballStemmer.java
 *
 *
 * Copyright (c) 2000-2001, The University of Sheffield.
 *
 * This file is part of GATE (see http://gate.ac.uk/), and is free
 * software, licenced under the GNU Library General Public License,
 * Version 2, June1991.
 *
 * A copy of this licence is included in the distribution in the file
 * licence.html, and is also available at http://gate.ac.uk/gate/licence.html.
 *
 *  milena, 26/3/2004
 *
 */


package stemmer;


import java.util.*;
import java.lang.*;
import gate.*;
import gate.creole.*;
import gate.util.*;
import java.lang.reflect.*;
import stemmer.snowball.SnowballProgram;
import gate.event.*;

 /**

  * Title: SnowballStemmer.java <br>
  * Description: This class is a wrapper for the
  * Snowball stemmers (see <a href = http://snowball.tartarus.org/> http://snowball.tartarus.org/</a>) for
  * <UL>
  * <Li>Danish, <Li>Dutch,
  * <Li>English, <Li>Finnish,
  * <Li>French, <Li>German,
  * <Li>Italian, <Li>Norwegian,
  * <Li>Portuguese, <Li>Russian,
  * <Li>Spanish, <Li>Swedish.</UL>
  * The Stemmer process already created annotations containing string feature,
  * as these created by the Tokeniser.
  * Generally it takes the (string) value form the parameter {@link annotationFeature}
  * of {@link annotationType} parameter
  * in {@link annotationSetName} parameter,
  * stems it and
  * adds new feature "stem" into the same annotation type. <br>
  * The default parameters are set to process Token.string values
  * form the Default annotation set, produced by the Tokeniser.
  * @author Milena Yankova
  * @version 1.0
 */


public class SnowballStemmer  extends AbstractLanguageAnalyser
    implements ProcessingResource {

/** Document to be processed by the stemmer - Runtime */
  public static final String
  SNOW_STAM_DOCUMENT_PARAMETER_NAME = "document";

/** Name of the annotation set the stemmer will run over  - Optional and  Runtime */
  public static final String
  SNOW_STAM_ANNOT_SET_PARAMETER_NAME = "annotationSetName";

/** Name of the annotation type in the {@link annotationSetName} to be processed - Runtime */
  public static final String
  SNOW_STAM_ANNOT_TYPE_PARAMETER_NAME = "annotationType";

/** Name of the feature containing a string of the word to be stemmed - Runtime */
  public static final String
  SNOW_STAM_ANNOT_FEATURE_PARAMETER_NAME = "annotationFeature";

/** Language that the stemmer will work on - Initialisation parameter */
  public static final String
  SNOW_STAM_LANGUAGE_PARAMETER_NAME = "language";

  private Class stemClass;
  private String language;
  private String annotationSetName;
  private String annotationType;
  private String annotationFeature;

  /** Default constructor*/
  public SnowballStemmer() {  }


   /** Initialises this resource, and returns it
    * <br>Checks if the language given as a parameter is supported.
    * @return Resource
    * @throws ResourceInstantiationException
    * */
   public Resource init() throws ResourceInstantiationException{
       fireStatusChanged("Creating a stemmer");
       FeatureMap params = Factory.newFeatureMap();
       fireProgressChanged(100);
       fireProcessFinished();

       /*report unsepported language choise*/
       if (language.equals("danish")||
           language.equals("dutch") ||
           language.equals("english")||
           language.equals("finnish")||
           language.equals("french")||
           language.equals("german")||
           language.equals("italian")||
           language.equals("norwegian")||
           language.equals("portuguese")||
           language.equals("russian")||
           language.equals("spanish")||
           language.equals("swedish")) {} else throw new ResourceInstantiationException("Unsupported language:  " +language);

       return this;

  }


  /** Runs the appropriate Snowball stemmer for the given language
   * @throws ExecutionException
   *  */
  public void execute() throws ExecutionException{

       interrupted = false;

       /*check the parameters*/
        if(document == null) throw new GateRuntimeException(
          "No document to process!");
        fireProgressChanged(0);
        fireStatusChanged("Stemming " + document.getName() + "...");

       /*get annotation set of all items with annotationType form the annotationSetName*/
       if(annotationSetName != null && annotationSetName.equals("")) annotationSetName = null;
       AnnotationSet inputAS = (annotationSetName == null) ?
                               document.getAnnotations() :
                               document.getAnnotations(annotationSetName);
       AnnotationSet tokensAS = inputAS.get(annotationType);

       if( tokensAS == null) throw new GateRuntimeException(
          "No annotations to process!\n" + "Please run Tokeniser first, if using default Stemmer features!");

       Object [] emptyArgs = new Object[0];
       Iterator iter = tokensAS.iterator();

       int allTokens = tokensAS.size();
       int processedTokens=0;
       int lastReport = 0;

       try{
         /*get the appropreate class and method according to the language */
         Class stemClass = Class.forName("stemmer.snowball.ext." + language + "Stemmer");
         SnowballProgram stemmer =  (SnowballProgram)stemClass.newInstance();
         Method stemMethod = stemClass.getMethod("stem", null);

         /*process each token separately*/
         while (iter.hasNext()){

           if(isInterrupted()) throw new ExecutionInterruptedException(
               "The execution of the \"" + getName() +
               "\" stemmer has been abruptly interrupted!");

           /*get the token string*/
           Annotation token = (Annotation)iter.next();
           FeatureMap allFeatures = token.getFeatures();
           String tokenString = (String)allFeatures.get(annotationFeature);
           //  if(allFeatures.get("kind").equals("word")){

           /*stem the token string*/
           stemmer.setCurrent(tokenString.toLowerCase());
           stemMethod.invoke(stemmer, emptyArgs);
           allFeatures.put("stem", (String)stemmer.getCurrent());
           token.setFeatures(allFeatures);

           processedTokens++;
           if(processedTokens - lastReport > 100){
             lastReport = processedTokens;
             fireProgressChanged(processedTokens * 100 / allTokens );
           }  //}

         }
         fireProcessFinished();

         }catch (Exception ce)
             {  throw new ExecutionException(ce);}

  }



  public void setLanguage(String language) {
     this.language = language;
   }
   public String getLanguage() {
     return language;
   }


   public void setAnnotationSetName(String annotationSetName) {
     this.annotationSetName = annotationSetName;
   }
   public String getAnnotationSetName() {
     return annotationSetName;
   }

   public void setAnnotationType(String annotationType) {
     this.annotationType = annotationType;
   }
   public String getAnnotationType() {
     return annotationType;
   }

   public void setAnnotationFeature(String annotationFeature) {
     this.annotationFeature = annotationFeature;
   }
   public String getAnnotationFeature() {
     return annotationFeature;
   }




} // class Stemmer








